<?php

namespace WPFunnelsPro\Frontend\Gateways;

use WPFunnelsPro\Wpfnl_Pro_functions;

/**
 * Class Wpfnl_Pro_Gateway_Mollie
 * @package WPFunnelsPro\Frontend\Gateways
 */
class Wpfnl_Pro_Gateway_Mollie extends Wpfnl_Pro_Mollie_Helper {

    /**
     * @var string
     */
    public $key = 'mollie_wc_gateway_creditcard';


    /**
     * @var bool
     */
    public $refund_support;


    public function __construct() {
        $this->refund_support = true;

        if( class_exists('Mollie_WC_Plugin') ) {
            add_filter( 'woocommerce_mollie_wc_gateway_creditcard_args', array( $this, 'create_non_logged_in_mollie_customer_id' ), 10, 2 );
        }

        add_action( 'wp_ajax_wpf_mollie_credit_card_payment_process', array( $this, 'process_credit_card' ) );
        add_action( 'wp_ajax_nopriv_wpf_mollie_credit_card_payment_process', array( $this, 'process_credit_card' ) );

        // run this action after successful payment to mollie
        add_action( 'woocommerce_api_wpf_mollie_cc_webhook', array( $this, 'maybe_process_mollie_webhook' ) );

        // hook for saving refund meta
        add_action( 'wpfunnels/child_order_created_' . $this->key, array( $this, 'save_meta_for_refund' ), 10, 3 );

        // hook for saving subscription meta
        add_action( 'wpfunnels/subscription_created', array( $this, 'save_offer_subscription_meta' ), 10, 3 );
    }


    /**
     * after payment successfully processed
     *
     * @param $order
     * @param $offer_product
     * @return bool
     */
    public function process_payment( $order, $offer_product ) {
        $result = array(
            'is_success' => false,
            'message' => ''
        );
        $tr_id = $order->get_meta( '_mollie_' . $offer_product['step_id'] . '_tr_id', true );
        update_post_meta( $order->get_id(), '_wpfunnels_offer_txn_resp_' . $offer_product['step_id'], $tr_id );

        if ( '' !== $tr_id ) {
            $result['is_success'] = true;
        }
        return $result;
    }


    /**
     * process ajax for mollie credit card
     */
    public function process_credit_card() {
        $nonce = filter_input( INPUT_POST, 'security', FILTER_SANITIZE_STRING );

        if ( ! wp_verify_nonce( $nonce, 'wpfnl_mollie_cc_process_nonce' ) ) {
            return;
        }
        $step_id       = isset( $_POST['step_id'] ) ? intval( $_POST['step_id'] ) : 0;
        $order_id      = isset( $_POST['order_id'] ) ? intval($_POST['order_id']) : 0;
        $offer_type    = isset( $_POST['offer_type'] ) ? sanitize_text_field( wp_unslash( $_POST['offer_type'] ) ) : '';
        $offer_action  = isset( $_POST['offer_action'] ) ? sanitize_text_field( wp_unslash( $_POST['offer_action'] ) ) : '';
        $product_id    = isset( $_POST['product_id'] ) ? intval( $_POST['product_id'] ) : '';
        $quantity      = isset( $_POST['quantity'] ) ? intval( $_POST['quantity'] ) : '';
        $order         = wc_get_order( $order_id );
        $offer_product = Wpfnl_Pro_functions::get_offer_product_data( $step_id, $product_id, 0, $order_id );
        $step_type     = get_post_meta( $step_id, '_step_type', true );

        if ( isset($offer_product['price']) && (floatval(0) === floatval( $offer_product['price'] ) || '' === trim($offer_product['price'])) ) {
            wp_send_json(array(
                'result'    => 'fail',
                'message'   => __('Product price is less than 0', 'wpfnl-pro'),
            ));
        } else {
            $mollie_settings    = \Mollie_WC_Plugin::getSettingsHelper();
            $is_test_mode       = $mollie_settings->isTestModeEnabled();
            $customer_id        = $this->get_mollie_user_id( $order, $is_test_mode );
            if( $customer_id ) {
                $mollie_api     = \Mollie_WC_Plugin::getApiHelper()->getApiClient( $is_test_mode );

                // set redirect url
                $redirect_url   = add_query_arg( array(
                    'wpfnl-order'           => $order->get_id(),
                    'wpfnl-key'             => $order->get_order_key(),
                    'wpfnl-mollie-return'   => true,
                ), get_permalink($step_id) );

                // set webhook url
                $webhook_url = add_query_arg( array(
                    'step_id'   => $step_id,
                    'order_id'  => $order_id,
                    'order_key' => $order->get_order_key(),
                ), WC()->api_request_url( 'wpf_mollie_cc_webhook' ) );


                $paymentLocale = $mollie_settings->getPaymentLocale();
                $payment_data   = array(
                    'amount' => array(
                        'currency'  => \Mollie_WC_Plugin::getDataHelper()->getOrderCurrency( $order ),
                        'value'     => \Mollie_WC_Plugin::getDataHelper()->formatCurrencyValue( $offer_product['price'], \Mollie_WC_Plugin::getDataHelper()->getOrderCurrency( $order ) )
                    ),
                    'description'   => "One click {$step_type} offer {$order_id}_{$step_id}",
                    'redirectUrl'   => $redirect_url,
                    'webhookUrl'    => $webhook_url,
                    'method'        => 'creditcard',
                    'customerId'    => $customer_id,
                    'locale'        => $paymentLocale,
                    'metadata'      => array(
                        'order_id' => $order_id,
                    ),
                );
                $payment = $mollie_api->payments->create( $payment_data );
                wp_send_json(
                    array(
                        'result'   => 'success',
                        'redirect' => $payment->getCheckoutUrl(),
                    )
                );
            }
        }
    }


    /**
     * process webhook from mollie
     *
     */
    public function maybe_process_mollie_webhook() {
        if( empty( $_GET['step_id'] ) || empty( $_GET['order_id'] ) || empty( $_GET['order_key'] ) ) {
            \Mollie_WC_Plugin::setHttpResponseCode( 400 );
            return;
        }

        $step_id       = intval( $_GET['step_id'] );
        $order_id      = intval( $_GET['order_id']);
        $order_key     = sanitize_text_field($_GET['order_key']);
        $order         = wc_get_order( $order_id );
        $step_type     = get_post_meta( $step_id, '_step_type', true );

        // if order not found
        if ( ! $order ) {
            \Mollie_WC_Plugin::setHttpResponseCode( 400 );
            return;
        }

        $test_mode  = $order->get_meta( '_mollie_payment_mode', true ) == 'test';
        $payment_id = sanitize_text_field( wp_unslash( $_POST['id'] ) );

        try {
            $payment_object = \Mollie_WC_Plugin::getPaymentFactoryHelper()->getPaymentObject( $payment_id );
        } catch ( \Mollie\Api\Exceptions\ApiException $exception ) {
            \Mollie_WC_Plugin::setHttpResponseCode( 400 );
            return;
        }

        $payment = $payment_object->getPaymentObject( $payment_object->data, $test_mode, $use_cache = false );

        // if payment not found
        if ( ! $payment ) {
            \Mollie_WC_Plugin::setHttpResponseCode( 404 );
            return;
        }

        // if order id and payment meta order id is not matched
        if ( $order_id != $payment->metadata->order_id ) {
            \Mollie_WC_Plugin::setHttpResponseCode( 400 );
            return;
        }

        $order->update_meta_data( '_mollie_' . $step_id . '_tr_id', $payment_id );
        $order->save();
        $order->add_order_note( __( "Payment processed for mollie credit card.", 'wpfnl-pro' ) );
    }

}