<?php
namespace WPFunnelsPro\Frontend\Gateways;

/**
 * @author [Saiduzzaman Tohin]
 * @email [tohin@coderex.co]
 * @create date 2022-05-26 11:16:34
 * @modify date 2022-05-26 11:16:34
 * @desc [Supporting Square Payment Gateway with WPFunnels one-click offer]
 */

use Square\SquareClient;
use WooCommerce\Square\Framework\PaymentGateway\Payment_Gateway_Helper;
use WooCommerce\Square\Framework\Square_Helper;
use WPFunnelsPro\Wpfnl_Pro_functions;
use WPFunnels\Wpfnl_functions;

class Wpfnl_Pro_Gateway_Square {

    /**
     * @var string
     * @since 1.6.4
     */
    public $key = 'square_credit_card';

    /**
     * @var bool
     * @since 1.6.4
     */
    public $refund_support;


    public function __construct() {

        $this->refund_support = true;
		add_filter( 'wc_' . $this->key . '_payment_form_tokenization_forced', array( $this, 'should_square_tokenization_allowed' ), 10, 2 );
		add_filter( 'wc_' . $this->key . '_payment_form_tokenization_allowed', array( $this, 'should_square_tokenization_allowed' ), 10, 2 );
        add_filter( 'wc_payment_gateway_' . $this->key . '_get_order', array( $this, 'square_should_create_token' ), 10, 2 );
        add_action( 'wpfunnels/child_order_created_' . $this->key, array( $this, 'add_capture_meta_to_child_order' ), 9999, 3 );
        add_action( 'wpfunnels/subscription_created', array( $this, 'add_offer_subscription_meta' ), 9999, 3 );

    }


    /**
     * Square should create payment token for offer product charge
     *
     * @param $order
     * @param $gateway
	 * 
     * @return bool|\WC_Order|\WC_Order_Refund
	 * @since 1.6.4
     */
    function square_should_create_token($order, $gateway) {
        if ( empty( $order->payment->token ) && $order->get_customer_id() < 1 ) {
            $order  = $this->get_order($order);
            $response = $gateway->get_api()->create_customer($order);
            $order->square_customer_id = $order->customer_id = $response->get_customer_id();
            $response = $gateway->get_payment_tokens_handler()->create_token($order);
        }
        return $order;
    }


    /**
	 * Remove underscore and add dash inside a string
	 * 
     * @return string|string[]
	 * @since 1.6.4
     */
    public function get_id_dasherized() {
        return str_replace( '_', '-', $this->key );
    }


    /**
     * Return order information for creating customer response
     *
     * @param $order
	 * 
     * @return bool|\WC_Order|\WC_Order_Refund
	 * @since 1.6.4
     */
    public function get_order( $order ) {

        if ( is_numeric( $order ) ) {
            $order = wc_get_order( $order );
        }

        $order->payment = new \stdClass();

        if ( empty( $order->payment->token ) ) {
            $order->payment->nonce          = Square_Helper::get_post( 'wc-' . $this->get_id_dasherized() . '-payment-nonce' );
            $order->payment->card_type      = Payment_Gateway_Helper::normalize_card_type( Square_Helper::get_post( 'wc-' . $this->get_id_dasherized() . '-card-type' ) );
            $order->payment->account_number = $order->payment->last_four = substr( Square_Helper::get_post( 'wc-' . $this->get_id_dasherized() . '-last-four' ), -4 );
            $order->payment->exp_month      = Square_Helper::get_post( 'wc-' . $this->get_id_dasherized() . '-exp-month' );
            $order->payment->exp_year       = Square_Helper::get_post( 'wc-' . $this->get_id_dasherized() . '-exp-year' );
            $order->payment->postcode       = Square_Helper::get_post( 'wc-' . $this->get_id_dasherized() . '-payment-postcode' );
        }
        return $order;
    }


    /**
	 * Payment Gateway Payment Form Tokenization Forced.
	 *
	 * Filters whether tokenization is forced for the payment form.
	 *
	 * @since 1.6.4
	 *
	 * @param bool $tokenization_forced
	 */
	public function should_square_tokenization_allowed($tokenization_forced)
	{
		if(get_the_ID() === 1){
			$post = Wpfnl_Pro_functions::get_sanitized_get_post();
			$checkout_id = Wpfnl_functions::get_checkout_id_from_post($post);
		}else{
			$checkout_id = get_the_ID();
		}
        $funnel_id   = Wpfnl_functions::get_funnel_id_from_step( $checkout_id );
        if ( $checkout_id && $funnel_id ) {
            $next_step  = Wpfnl_functions::get_next_step($funnel_id, $checkout_id);

            if ( $next_step && Wpfnl_Pro_functions::is_offer_page($next_step['step_id']) ) {
                $tokenization_forced = true;
            }
        }
        return $tokenization_forced;
	}


    /**
     * Processing of the Upsell or Downsell offers
     * 
     * @param mixed $order
     * @param mixed $offer_product
     * 
     * @return array 
     * @since 1.6.4
     */
    public function process_payment($order, $offer_product) {
		$currency = $order->get_currency();
		// Setup square payment configuration 
        $client = $this->set_square_client_environment();

        if ( isset($offer_product['price']) && (floatval(0) === floatval( $offer_product['price'] ) || '' === trim($offer_product['price'])) ) {
            wp_send_json(array(
                'result'    => 'fail',
                'message'   => __('Product price is less than 0', 'wpfnl-pro'),
            ));
        } else {
            $money_utility_object = new \WooCommerce\Square\Utilities\Money_Utility();
            $amount_money = $money_utility_object::amount_to_money($offer_product['total'], $currency);
            
            $credit_card_payment_token = get_post_meta($order->get_id(), '_wc_square_credit_card_payment_token', true);
            $credit_card_customer_id = get_post_meta($order->get_id(), '_wc_square_credit_card_customer_id', true);

			// Prepare data structure for request
            $body = new \Square\Models\CreatePaymentRequest(
                $credit_card_payment_token,
                uniqid(),
                $amount_money
            );
            $body->setAutocomplete(true);
            $body->setCustomerId($credit_card_customer_id);

			// Handle payment request API responses
            $api_response = $client->getPaymentsApi()->createPayment($body);

            if ($api_response->isSuccess()) {
				$transaction_id = $api_response->getResult()->getPayment()->getId();

				// Store transaction ID for offer products
				$this->store_offer_transaction( $order, $transaction_id, $offer_product );
				// Reduce offer products from stock
				$this->reduce_the_offer_product_from_stock($order, $offer_product);

                return array(
                    'is_success' => true,
                    'message' => 'Success'
                );
            } else {
				$errors = $api_response->getErrors();
                return array(
					'is_success' => false,
					'message' => sprintf( __( 'Square payment transaction failed (%1$s:%2$s)', 'wpfnl-pro' ), $errors )
				);
            }
                        
        }
    }


	/**
	 * Store transaction ID for offer products
	 * 
     * @param $order
     * @param $response
     * @param $product
	 * 
	 * @since 1.6.4
     */
    public function store_offer_transaction( $order, $transaction_id, $product )
    {
        $order->update_meta_data('_wpfunnels_offer_txn_resp_' . $product['step_id'], $transaction_id);
        $order->save();
    }


	/**
	 * Set up the required configuration for payment.
	 * 
	 * @return object
	 * 
	 * @since 1.6.4
	 */
	public function set_square_client_environment() {

		// Set the access tokan.
		$this->access_token = $this->get_wc_gateway()->get_plugin()->get_settings_handler()->get_access_token();
		$this->access_token = empty( $this->access_token ) ? $this->get_wc_gateway()->get_plugin()->get_settings_handler()->get_option( 'sandbox_token' ) : $this->access_token;

		// Set the location id.
		$this->location_id = $this->get_wc_gateway()->get_plugin()->get_settings_handler()->get_location_id();
		$client = new SquareClient([
            'accessToken' => $this->access_token,
            'environment' => $this->get_environment(),
        ]);

		return $client;
	}


	/**
	 * Get WooCommerce payment geteways.
	 *
	 * @return array
	 * 
	 * @since 1.6.4
	 */
	public function get_wc_gateway() {

		global $woocommerce;

		$gateways = $woocommerce->payment_gateways->payment_gateways();

		return $gateways[ $this->key ];
	}


	/**
	 * Gets the configured environment.
	 *
	 * @since 1.6.4
	 *
	 * @return string
	 */
	public function get_environment() {
		$sanboxed = ( defined( 'WC_SQUARE_SANDBOX' ) && WC_SQUARE_SANDBOX ) || $this->is_sandbox_setting_enabled();
		return $sanboxed ? 'sandbox' : 'production';
	}


	/**
	 * Tells is if the setting for enabling sandbox is checked.
	 *
	 * @since 1.6.4
	 *
	 * @return boolean
	 */
	public function is_sandbox_setting_enabled() {
		return 'yes' === $this->get_enable_sandbox();
	}


	/**
	 * Gets setting enabled sandbox.
	 *
	 * @since 1.6.4
	 *
	 * @return string
	 */
	public function get_enable_sandbox() {
        $setting = get_option('wc_square_settings');
		return $setting['enable_sandbox'];
	}


	/**
	 * Determines if configured in the sandbox environment.
	 *
	 * @since 1.6.4
	 *
	 * @return bool
	 */
	public function is_sandbox() {

		return 'sandbox' === $this->get_environment();
	}


    /**
     * Update post meta for refund process on upsell or downsell
     *
     * @param $parent_order
     * @param $child_order
     * @param $transaction_id
     * 
     * @since 1.6.4
     */
    public function add_capture_meta_to_child_order( $parent_order, $child_order, $transaction_id ) {
        update_post_meta( $child_order->get_id(), '_wc_square_credit_card_trans_id', $parent_order->data['transaction_id'] );
		update_post_meta( $child_order->get_id(), '_wc_square_credit_card_square_location_id', $parent_order->get_meta( '_wc_square_credit_card_square_location_id', true ) );
		update_post_meta( $child_order->get_id(), '_wc_square_credit_card_authorization_code', $parent_order->data['transaction_id'] );
		update_post_meta( $child_order->get_id(), '_wc_square_credit_card_customer_id', $parent_order->get_meta( '_wc_square_credit_card_customer_id', true ) );
		update_post_meta( $child_order->get_id(), '_wc_square_credit_card_square_order_id', $parent_order->data['transaction_id'] );
		update_post_meta( $child_order->get_id(), '_wc_square_credit_card_charge_captured', 'yes' );
    }


	/**
	 * Process offer refund
	 *
	 * @param object $order Order Object.
	 * @param array  $offer_data offer data.
	 *
	 * @return string/bool.
	 * 
	 * @since 1.6.4
	 */
	public function process_refund_offer( $order, $offer_data ) {
		$payment_id = $offer_data['transaction_id'];

		// set up the square payment configuration.
		$client = $this->set_square_client_environment();
		$money_utility_object = new \WooCommerce\Square\Utilities\Money_Utility();
        $amount_money = $money_utility_object::amount_to_money($offer_data['amount'],'USD');

		$body = new \Square\Models\RefundPaymentRequest(
			uniqid(), 
			$amount_money,
			$payment_id
		);

		$api_response = $client->getRefundsApi()->refundPayment($body);
	
		if ($api_response->isSuccess()) {
			$response_id = $api_response->getResult()->getRefund()->getId();
			return $response_id;
		} else {
			$errors = $api_response->getErrors();
			return false;
		}

	}


	/**
     * add subscription offer meta to order
     *
     * @param $subscription
     * @param $offer_product
     * @param $order
	 * 
	 * @since 1.6.4
     */
    public function add_offer_subscription_meta( $subscription, $offer_product, $order ) {
        if ( $this->key === $order->get_payment_method() ) {
            $subscription_id = $subscription->get_id();
            update_post_meta( $subscription_id, '_wc_square_credit_card_payment_token', $order->get_meta( '_wc_square_credit_card_payment_token', true ) );
            update_post_meta( $subscription_id, '_wc_square_credit_card_customer_id', $order->get_meta( '_wc_square_credit_card_customer_id', true ) );
        }
    }



	/**
	 * Reduce the offer product from the stock after upsell or downsell offer has been accepted
	 *
	 * @param object $order Object of order.
	 * @param array  $offer_product array of offer product.
	 * 
	 * @since 1.6.4
	 */
	public function reduce_the_offer_product_from_stock( $order, $offer_product ) {

		$product = wc_get_product( $offer_product['id'] );
		if( $product ){
			$new_stock = wc_update_product_stock( $offer_product['id'], $offer_product['qty'], 'decrease' );
			$changes[] = array(
				'product' => $product,
				'from'    => $new_stock + intval( $offer_product['qty'] ),
				'to'      => $new_stock,
			);
			wc_trigger_stock_change_notifications( $order, $changes );
		}
		

	}

}